MODULE dynnxt
   !!======================================================================
   !!                       ***  MODULE  dynnxt  ***
   !! Ocean dynamics: time stepping
   !!======================================================================
   
   !!----------------------------------------------------------------------
   !!   dyn_nxt      : update the horizontal velocity from the momentum trend
   !!----------------------------------------------------------------------
   !! * Modules used
   USE oce             ! ocean dynamics and tracers
   USE dom_oce         ! ocean space and time domain
   USE in_out_manager  ! I/O manager
   USE obc_oce         ! ocean open boundary conditions
   USE obcdyn          ! open boundary condition for momentum (obc_dyn routine)
   USE obcdyn_bt       ! 2D open boundary condition for momentum (obc_dyn_bt routine)
   USE obcvol          ! ocean open boundary condition (obc_vol routines)
   USE bdy_oce         ! unstructured open boundary conditions
   USE bdydta          ! unstructured open boundary conditions
   USE bdydyn          ! unstructured open boundary conditions
   USE dynspg_oce      ! type of surface pressure gradient
   USE lbclnk          ! lateral boundary condition (or mpp link)
   USE prtctl          ! Print control
   USE agrif_opa_update
   USE agrif_opa_interp
   USE domvvl          ! variable volume

   IMPLICIT NONE
   PRIVATE

   !! * Accessibility
   PUBLIC dyn_nxt                ! routine called by step.F90
   !! * Substitutions
#  include "domzgr_substitute.h90"
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE dyn_nxt ( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dyn_nxt  ***
      !!                   
      !! ** Purpose :   Compute the after horizontal velocity from the 
      !!      momentum trend.
      !!
      !! ** Method  :   Apply lateral boundary conditions on the trends (ua,va) 
      !!      through calls to routine lbc_lnk.
      !!      After velocity is compute using a leap-frog scheme environment:
      !!         (ua,va) = (ub,vb) + 2 rdt (ua,va)
      !!      Note that if lk_dynspg_flt=T, the time stepping has already been
      !!      performed in dynspg module
      !!      Time filter applied on now horizontal velocity to avoid the
      !!      divergence of two consecutive time-steps and swap of dynamics
      !!      arrays to start the next time step:
      !!         (ub,vb) = (un,vn) + atfp [ (ub,vb) + (ua,va) - 2 (un,vn) ]
      !!         (un,vn) = (ua,va) 
      !!
      !! ** Action : - Update ub,vb arrays, the before horizontal velocity
      !!             - Update un,vn arrays, the now horizontal velocity
      !!
      !! History :
      !!        !  87-02  (P. Andrich, D. L Hostis)  Original code
      !!        !  90-10  (C. Levy, G. Madec)
      !!        !  93-03  (M. Guyon)  symetrical conditions
      !!        !  97-02  (G. Madec & M. Imbard)  opa, release 8.0
      !!        !  97-04  (A. Weaver)  Euler forward step
      !!        !  97-06  (G. Madec)  lateral boudary cond., lbc routine
      !!   8.5  !  02-08  (G. Madec)  F90: Free form and module
      !!        !  02-10  (C. Talandier, A-M. Treguier) Open boundary cond.
      !!   9.0  !  05-11  (V. Garnier) Surface pressure gradient organization
      !!    "   !  07-07  (D. Storkey) Calls to BDY routines. 
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt      ! ocean time-step index

      !! * Local declarations
      INTEGER  ::   ji, jj, jk   ! dummy loop indices
      REAL(wp) ::   z2dt         ! temporary scalar
      REAL(wp) ::   zsshun1, zsshvn1         ! temporary scalar
      !! Variable volume
      REAL(wp), DIMENSION(jpi,jpj)     ::  & ! 2D workspace
         zsshub, zsshun, zsshua,           &
         zsshvb, zsshvn, zsshva
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::  &
         zfse3ub, zfse3un, zfse3ua, &        ! 3D workspace
         zfse3vb, zfse3vn, zfse3va
      !!----------------------------------------------------------------------
      !!  OPA 9.0 , LOCEAN-IPSL (2005) 
      !! $Header: /home/opalod/NEMOCVSROOT/NEMO/OPA_SRC/DYN/dynnxt.F90,v 1.13 2007/05/25 15:51:50 opalod Exp $ 
      !! This software is governed by the CeCILL licence see modipsl/doc/NEMO_CeCILL.txt 
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn_nxt : time stepping'
         IF(lwp) WRITE(numout,*) '~~~~~~~'
      ENDIF

      ! Local constant initialization
      z2dt = 2. * rdt
      IF( neuler == 0 .AND. kt == nit000 )  z2dt = rdt

      !! Explicit physics with thickness weighted updates
      IF( lk_vvl .AND. .NOT. lk_dynspg_flt ) THEN

         ! Sea surface elevation time stepping
         ! -----------------------------------
         !
         DO jj = 1, jpjm1
            DO ji = 1,jpim1

               ! Sea Surface Height at u-point before
               zsshub(ji,jj) = 0.5 * umask(ji,jj,1) / ( e1u(ji,jj) * e2u(ji,jj) )     &
                  &          * ( e1t(ji  ,jj  ) * e2t(ji  ,jj  ) * sshbb(ji  ,jj  )   &
                  &            + e1t(ji+1,jj  ) * e2t(ji+1,jj  ) * sshbb(ji+1,jj  ) )

               ! Sea Surface Height at v-point before
               zsshvb(ji,jj) = 0.5 * vmask(ji,jj,1) / ( e1v(ji,jj) * e2v(ji,jj) )     &
                  &          * ( e1t(ji  ,jj  ) * e2t(ji  ,jj  ) * sshbb(ji  ,jj  )   &
                  &            + e1t(ji  ,jj+1) * e2t(ji  ,jj+1) * sshbb(ji  ,jj+1) )

               ! Sea Surface Height at u-point after
               zsshua(ji,jj) = 0.5 * umask(ji,jj,1) / ( e1u(ji,jj) * e2u(ji,jj) )     &
                  &          * ( e1t(ji  ,jj  ) * e2t(ji  ,jj  ) * ssha(ji  ,jj  )    &
                  &            + e1t(ji+1,jj  ) * e2t(ji+1,jj  ) * ssha(ji+1,jj  ) )

               ! Sea Surface Height at v-point after
               zsshva(ji,jj) = 0.5 * vmask(ji,jj,1) / ( e1v(ji,jj) * e2v(ji,jj) )     &
                  &          * ( e1t(ji  ,jj  ) * e2t(ji  ,jj  ) * ssha(ji  ,jj  )    &
                  &            + e1t(ji  ,jj+1) * e2t(ji  ,jj+1) * ssha(ji  ,jj+1) )

            END DO
         END DO

         ! Boundaries conditions
         CALL lbc_lnk( zsshua, 'U', 1. )   ;    CALL lbc_lnk( zsshva, 'V', 1. )
         CALL lbc_lnk( zsshub, 'U', 1. )   ;    CALL lbc_lnk( zsshvb, 'V', 1. )

         ! Scale factors at before and after time step
         ! -------------------------------------------
         CALL dom_vvl_sf( zsshub, 'U', zfse3ub ) ;    CALL dom_vvl_sf( zsshua, 'U', zfse3ua ) 
         CALL dom_vvl_sf( zsshvb, 'V', zfse3vb ) ;    CALL dom_vvl_sf( zsshva, 'V', zfse3va ) 

         ! Asselin filtered scale factor at now time step
         ! ----------------------------------------------
         IF( (neuler == 0 .AND. kt == nit000) .OR. lk_dynspg_ts ) THEN
            CALL dom_vvl_sf_ini( 'U', zfse3un ) ;   CALL dom_vvl_sf_ini( 'V', zfse3vn ) 
         ELSE
            zsshun(:,:) = atfp * ( zsshub(:,:) + zsshua(:,:) ) + atfp1 * sshu(:,:)
            zsshvn(:,:) = atfp * ( zsshvb(:,:) + zsshva(:,:) ) + atfp1 * sshv(:,:)
            CALL dom_vvl_sf( zsshun, 'U', zfse3un ) ;   CALL dom_vvl_sf( zsshvn, 'V', zfse3vn ) 
         ENDIF

         ! Thickness weighting
         ! -------------------
         DO jk = 1, jpkm1
            DO jj = 1, jpj
               DO ji = 1, jpi
                  ua(ji,jj,jk) = ua(ji,jj,jk) * fse3u(ji,jj,jk)
                  va(ji,jj,jk) = va(ji,jj,jk) * fse3v(ji,jj,jk)

                  un(ji,jj,jk) = un(ji,jj,jk) * fse3u(ji,jj,jk)
                  vn(ji,jj,jk) = vn(ji,jj,jk) * fse3v(ji,jj,jk)

                  ub(ji,jj,jk) = ub(ji,jj,jk) * zfse3ub(ji,jj,jk)
                  vb(ji,jj,jk) = vb(ji,jj,jk) * zfse3vb(ji,jj,jk)
               END DO
            END DO
         END DO

      ENDIF

      ! Lateral boundary conditions on ( ua, va )
      CALL lbc_lnk( ua, 'U', -1. )
      CALL lbc_lnk( va, 'V', -1. )

      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
         ! Next velocity
         ! -------------
#if defined key_dynspg_flt
         ! Leap-frog time stepping already done in dynspg.F routine
#else
         DO jj = 1, jpj                      ! caution: don't use (:,:) for this loop
            DO ji = 1, jpi                   ! it causes optimization problems on NEC in auto-tasking
               ! Leap-frog time stepping
               ua(ji,jj,jk) = ( ub(ji,jj,jk) + z2dt * ua(ji,jj,jk) ) * umask(ji,jj,jk)
               va(ji,jj,jk) = ( vb(ji,jj,jk) + z2dt * va(ji,jj,jk) ) * vmask(ji,jj,jk)
            END DO
         END DO

         IF( lk_vvl ) THEN
            ! Unweight velocities prior to updating open boundaries.

            DO jj = 1, jpj                      ! caution: don't use (:,:) for this loop
               DO ji = 1, jpi                   ! it causes optimization problems on NEC in auto-tasking
                  ua(ji,jj,jk) = ua(ji,jj,jk) / fse3u(ji,jj,jk)
                  va(ji,jj,jk) = va(ji,jj,jk) / fse3v(ji,jj,jk)

                  un(ji,jj,jk) = un(ji,jj,jk) / fse3u(ji,jj,jk)
                  vn(ji,jj,jk) = vn(ji,jj,jk) / fse3v(ji,jj,jk)

                  ub(ji,jj,jk) = ub(ji,jj,jk) / zfse3ub(ji,jj,jk)
                  vb(ji,jj,jk) = vb(ji,jj,jk) / zfse3vb(ji,jj,jk)
               END DO
            END DO

         ENDIF

# if defined key_obc
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============
      ! Update (ua,va) along open boundaries (only in the rigid-lid case)
      CALL obc_dyn( kt )

      IF ( lk_dynspg_exp .OR. lk_dynspg_ts ) THEN
         !Flather boundary condition :
         !        - Update sea surface height on each open boundary
         !                 sshn (= after ssh) for explicit case
         !                 sshn_b (= after ssha_b) for time-splitting case
         !        - Correct the barotropic velocities
         CALL obc_dyn_bt( kt )

         !Boundary conditions on sshn ( after ssh)
         CALL lbc_lnk( sshn, 'T', 1. )

         IF(ln_ctl) THEN         ! print sum trends (used for debugging)
            CALL prt_ctl(tab2d_1=sshn, clinfo1=' ssh      : ', mask1=tmask)
         ENDIF

         IF ( ln_vol_cst ) CALL obc_vol( kt )

      ENDIF

      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
# elif defined key_bdy 
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============
      ! Update (ua,va) along open boundaries (for exp or ts options).
      IF ( lk_dynspg_exp .or. lk_dynspg_ts ) THEN
  
         CALL bdy_dyn_frs( kt )

         IF ( ln_bdy_fla ) THEN

            ua_e(:,:)=0.0
            va_e(:,:)=0.0

            ! Set these variables for use in bdy_dyn_fla
            hu_e(:,:) = hu(:,:)
            hv_e(:,:) = hv(:,:)

            DO jk = 1, jpkm1
               !! Vertically integrated momentum trends
               ua_e(:,:) = ua_e(:,:) + fse3u(:,:,jk) * umask(:,:,jk) * ua(:,:,jk)
               va_e(:,:) = va_e(:,:) + fse3v(:,:,jk) * vmask(:,:,jk) * va(:,:,jk)
            END DO

            DO jk = 1 , jpkm1
               ua(:,:,jk) = ua(:,:,jk) - ua_e(:,:) * hur(:,:)
               va(:,:,jk) = va(:,:,jk) - va_e(:,:) * hvr(:,:)
            END DO

            CALL bdy_dta_bt( kt+1, 0)
            CALL bdy_dyn_fla

         ENDIF

         DO jk = 1 , jpkm1
            ua(:,:,jk) = ua(:,:,jk) + ua_e(:,:) * hur(:,:)
            va(:,:,jk) = va(:,:,jk) + va_e(:,:) * hvr(:,:)
         END DO

      ENDIF

      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
# endif
# if defined key_agrif
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============
      ! Update (ua,va) along open boundaries (only in the rigid-lid case)
      CALL Agrif_dyn( kt )
      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
# endif
#endif

         ! Time filter and swap of dynamics arrays
         ! ------------------------------------------
         IF( neuler == 0 .AND. kt == nit000 ) THEN
            IF( (lk_vvl .AND. .NOT. lk_dynspg_flt) ) THEN      ! Varying levels
               ! caution: don't use (:,:) for this loop
               ! it causes optimization problems on NEC in auto-tasking
               DO jj = 1, jpj
                  DO ji = 1, jpi
                     zsshun1 = umask(ji,jj,jk) / fse3u(ji,jj,jk)
                     zsshvn1 = vmask(ji,jj,jk) / fse3v(ji,jj,jk)
                     ub(ji,jj,jk) = un(ji,jj,jk) * zsshun1 * umask(ji,jj,jk)
                     vb(ji,jj,jk) = vn(ji,jj,jk) * zsshvn1 * vmask(ji,jj,jk)
                     zsshun1 = umask(ji,jj,jk) / zfse3ua(ji,jj,jk)
                     zsshvn1 = vmask(ji,jj,jk) / zfse3va(ji,jj,jk)
                     un(ji,jj,jk) = ua(ji,jj,jk) * zsshun1 * umask(ji,jj,jk)
                     vn(ji,jj,jk) = va(ji,jj,jk) * zsshvn1 * vmask(ji,jj,jk)
                  END DO
               END DO
            ELSE                                               ! Fixed levels
               DO jj = 1, jpj
                  DO ji = 1, jpi
                     ! Euler (forward) time stepping
                     ub(ji,jj,jk) = un(ji,jj,jk)
                     vb(ji,jj,jk) = vn(ji,jj,jk)
                     un(ji,jj,jk) = ua(ji,jj,jk)
                     vn(ji,jj,jk) = va(ji,jj,jk)
                  END DO
               END DO
            ENDIF
         ELSE
            IF( (lk_vvl .AND. .NOT. lk_dynspg_flt) ) THEN      ! Varying levels
               ! caution: don't use (:,:) for this loop
               ! it causes optimization problems on NEC in auto-tasking
               DO jj = 1, jpj
                  DO ji = 1, jpi
                     zsshun1 = umask(ji,jj,jk) / zfse3un(ji,jj,jk)
                     zsshvn1 = vmask(ji,jj,jk) / zfse3vn(ji,jj,jk)
                     ub(ji,jj,jk) = ( atfp  * ( ub(ji,jj,jk) + ua(ji,jj,jk) ) &
                       &            + atfp1 * un(ji,jj,jk) ) * zsshun1
                     vb(ji,jj,jk) = ( atfp  * ( vb(ji,jj,jk) + va(ji,jj,jk) ) &
                       &            + atfp1 * vn(ji,jj,jk) ) * zsshvn1
                     zsshun1 = umask(ji,jj,jk) / zfse3ua(ji,jj,jk)
                     zsshvn1 = vmask(ji,jj,jk) / zfse3va(ji,jj,jk)
                     un(ji,jj,jk) = ua(ji,jj,jk) * zsshun1
                     vn(ji,jj,jk) = va(ji,jj,jk) * zsshvn1
                  END DO
               END DO
            ELSE                                               ! Fixed levels
               DO jj = 1, jpj
                  DO ji = 1, jpi
                     ! Leap-frog time stepping
                     ub(ji,jj,jk) = atfp  * ( ub(ji,jj,jk) + ua(ji,jj,jk) ) + atfp1 * un(ji,jj,jk)
                     vb(ji,jj,jk) = atfp  * ( vb(ji,jj,jk) + va(ji,jj,jk) ) + atfp1 * vn(ji,jj,jk)
                     un(ji,jj,jk) = ua(ji,jj,jk)
                     vn(ji,jj,jk) = va(ji,jj,jk)
                  END DO
               END DO
            ENDIF
         ENDIF
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============

      IF(ln_ctl)   THEN
         CALL prt_ctl(tab3d_1=un, clinfo1=' nxt  - Un: ', mask1=umask, &
            &         tab3d_2=vn, clinfo2=' Vn: ', mask2=vmask)
      ENDIF

#if defined key_agrif
      IF (.NOT.Agrif_Root())    CALL Agrif_Update_Dyn( kt )
#endif      

   END SUBROUTINE dyn_nxt

   !!======================================================================
END MODULE dynnxt
